/* $Header: /cvsroot/sid/sid/user/util.c,v 1.17 2003/10/05 09:16:47 deppeler Exp $ */

#include "c_tsw.h"
#include "sha1_kernel.h"
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <math.h>

extern struct config_t *config;
extern int debug;

/* setprogname: set stored name of program */
void setprogname(char *str)
{
	config->name = estrdup(str);
}

/* estrdup: duplicate a string, report if error */
char *estrdup(char *s)
{
	char *t;

	t = (char *) malloc(strlen(s)+1);
	if (t == NULL)
		eprintf("estrdup(\"%.20s\") failed:", s);
	strcpy(t, s);
	return t;
}

/* progname: return stored name of program */
char * progname(void)
{
	return config->name;
}

/* eprintf: print error message and exit */
void eprintf(char *fmt, ...)
{
	va_list args;

	fflush(stdout);
	if (progname() != NULL)
		fprintf(stderr, "%s: ", progname());

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	if (fmt[0] != '\0' && fmt[strlen(fmt)-1] == ':')
		fprintf(stderr, " %s", strerror(errno));
	fprintf(stderr, "\n");
	exit(2); /* conventional value for failed execution */
}

/* debug printing */
void dprintf(int l, char *fmt, ...)
{
	va_list args;

	if ( debug < l )
		return;

	fflush(stdout);
	if (progname() != NULL)
		fprintf(stderr, "%s [debug]: ", progname());

	for ( ; l > 0; l--)
		fprintf(stderr, " ");

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	if (fmt[0] != '\0' && fmt[strlen(fmt)-1] == ':')
		fprintf(stderr, " %s", strerror(errno));
	fprintf(stderr, "\n");
}

/* free host_cfg_t list */
int free_host_cfg(host_cfg_t *c)
{
	host_cfg_t *next;
	tsw_time_t *next_automated_time;
	tsw_cmd_t *next_allowed_command_set;

	for( ; c != NULL; c = next ) {

		/* free automated_time list */
		for( ; c->automated_time != NULL; c->automated_time = next_automated_time ) {
			next_automated_time = c->automated_time->next;
			free(c->automated_time);
		}
		/* free command set list */
		for( ; c->allowed_command_set != NULL; c->allowed_command_set = next_allowed_command_set ) {
			next_allowed_command_set = c->allowed_command_set->next;
			free(c->allowed_command_set);
		}

		next = c->next;
		free(c);
	}
	return 0;
}

/* free cmd_def_t list */
int free_cmd_def(cmd_def_t *command_def)
{
	cmd_def_t *next;

	for( ; command_def != NULL; command_def = next ) {
		next = command_def->next;
		free(command_def);
	}
	return 0;
}

/* free config_t struct */
int free_all(config_t *c)
{
	free_host_cfg(c->host);
	free_cmd_def(c->command);
	free(c->name);
	free(c);
	return 0;
}

/* allocate new command definition */
cmd_def_t *new_cmd_def(void)
{
	cmd_def_t *newp;
	newp = (cmd_def_t *) ecalloc(1, sizeof(cmd_def_t));
	newp->next = NULL;
	return newp;
}

/* allocate new command */
tsw_cmd_t *new_cmd(void)
{
	tsw_cmd_t *newp;
	/*dprintf(7, "new_command");*/
	newp = (tsw_cmd_t *) ecalloc(1, sizeof(tsw_cmd_t));
	newp->prev = NULL;
	newp->next = NULL;
	return newp;
}

/* allocate new time specification */
tsw_time_t *new_time(void)
{
	tsw_time_t *newp;

	newp = (tsw_time_t *) ecalloc(1, sizeof(tsw_time_t));
	newp->next = NULL;
	return newp;
}


/* add new command definition in front (left) of the list */
cmd_def_t * add_cmd_def_l(cmd_def_t *listp, cmd_def_t *newp)
{
	newp->next = listp;
	return newp;
}

/* add new command in front (left) of the list */
tsw_cmd_t * add_cmd_l(tsw_cmd_t *listp, tsw_cmd_t *newp)
{
	newp->next = listp;
	if ( listp != NULL )
		listp->prev = newp;
	return newp;
}

/* add_time_front: add newp in fron of listp */
tsw_time_t * add_time_front(tsw_time_t *listp, tsw_time_t *newp)
{
	newp->next = listp;
	return newp;
}

/* addfront: add newp to front of listp */
host_cfg_t *addfront(host_cfg_t *listp, host_cfg_t *newp)
{
	newp->next = listp;
	return newp;
}

/* ecalloc: malloc and report if error */
void *ecalloc(size_t nelem, size_t elsize) 
{
	void *p;

	p = calloc(nelem, elsize);
	if (p == NULL)
		eprintf("malloc of %u bytes failed:", nelem*elsize);
	return p;
}

/* allocate new host configuration */
host_cfg_t *new_host_cfg(void)
{
	host_cfg_t *newp;

	newp = (host_cfg_t *) ecalloc(1, sizeof(host_cfg_t));
	newp->next = NULL;
	return newp;
}

/* allocate new state */
state_t *new_state(void)
{
	state_t *newp;

	newp = (state_t *) ecalloc(1, sizeof(state_t));
	newp->next = NULL;
	return newp;
}

/* add state in front (left) of list */
state_t *add_state_l(state_t *listp, state_t *newp)
{
	newp->next = listp;
	return newp;
}

/* add new state to the state list */
state_t *add_state(state_t *s, char *host, int pty) 
{
	dprintf(7, "add_state: tty: %d", pty);
	dprintf(7, "add_state: host: %s", host);
	dprintf(7, "add_state: state: %lx", s);
	s = add_state_l(s, new_state());
	s->host = estrdup(host);
	s->pty  = pty;
	s->hashes = 0;
	s->failures    = 0;
	s->automated_action    = 0;
	s->next_expected_hash = NULL;
	s->maintenance = FALSE;

	return s;
}

/* remove state from the state list */
state_t *rem_state(state_t *state, char *host, int pty) 
{
	state_t *p, *prev;

	prev = NULL;
	for ( p = state; p != NULL; p = p->next ) {
		dprintf(7, "rem_state: state: %lx", p);
		if ( strncmp(host, p->host, MAX_LINE_LENGTH) == 0 &&
				pty == p->pty ) {
			if ( prev == NULL )
				state = p->next;
			else
				prev->next = p->next;
			free(p);
			return state;
		}
		prev = p;
	}
	/* this happens when the parsing program is started after a pty
	has been opened and has never been allocated a state. */
	/* TODO: what to do when not in list */

	return state; 
}

/* get pointer to the state describing (host, pty) */
state_t *get_state(state_t *state, char *host, int pty) 
{
	state_t *r = NULL;

	for ( r = state; r != NULL; r = r->next ) {
		if ( strncmp(host, r->host, MAX_LINE_LENGTH) == 0 &&
				pty == r->pty ) {
			return r;
		}
	}
	return r;
}

/* get pointer to the host configuration describing the the host 'name' */
host_cfg_t *get_host(host_cfg_t *host, char *name)
{
	host_cfg_t *h = NULL;
	
	for ( h = host; h != NULL; h = h->next ) {
		dprintf(7, "get_host: checking host config at '%lx'", h);
		if ( strncmp(name, h->name, MAX_LINE_LENGTH) == 0 )
			return h;
	}
	return NULL;
}

/* check whether we're in maintenance or are going into maintenance */
int check_maintenance(config_t *config, state_t *s, char *hash)
{
	host_cfg_t *h = NULL;

	h = get_host(config->host, s->host);
	if ( h != NULL ) {
		dprintf(7, "check_maintenance: host found");
		if ( h->admin_mode == NONE )
			return 0;
		if ( s->maintenance == TRUE )
			return 1;

		if ( h->admin_mode == INTERACTIVE ) {
			if ( cmp_hash(hash, h->admin_digest) ) {
				dprintf(8, "check_maintenance: found host admin mode");
				s->maintenance = TRUE;
				return 1;
			} else {
				dprintf(8, "check_maintenance: not admin mode");
			}
		} else if ( h->admin_mode == GLOBAL ) {
			if ( cmp_hash(hash, config->admin_digest) ) {
				dprintf(8, "check_maintenance: found global admin mode");
				s->maintenance = TRUE;
				return 1;
			} else {
				dprintf(8, "check_maintenance: not admin mode");
			}
		}
	} else
		dprintf(7, "check_maintenance: no host found");

	s->failures++;
	return 0;
}

/* check whether the specified hash is a valid hash */
int check_hash(config_t *config, state_t *state, char *hash)
{
	cmd_info_t *cmd_info;

	dprintf(8, "check_hash: checking hash %s", hash);

	/* do we have any such hash/command configured? */
	cmd_info = find_cmd_info(config, state, hash);
	if ( cmd_info == NULL ) {
		dprintf(8, "check_hash: command definition not found");
		return 0;
	}

	/* are we within time specs for the found command definition */
	if ( time_ok(cmd_info->cmd_def->time) == 0 ) {
		dprintf(7, "check_hash: time out of spec");
		return 0;
	}
	dprintf(7, "check_hash: time ok");

	if ( state->hashes == 0 )
		state->in_sequence = cmd_info->cmd_def->ordered;

	dprintf(7, "check_hash: state->in_sequence: %d", state->in_sequence);
	if ( state->in_sequence == TRUE ) {
		if ( state->hashes == 0 ) {
			dprintf(9, "check_hash: 1st hash in state: setting next_expected_hash");
			state->next_expected_hash = cmd_info->command->digest;
		}
		if ( cmp_hash(hash, state->next_expected_hash) == 0 ) {
			dprintf(7, "check_hash: command is out of sequence");
			return 0;
		}
		dprintf(7, "command is in sequence");
		if ( cmd_info->command->prev != NULL ) {
			dprintf(9, "setting next_expected_hash to %s", cmd_info->command->prev->command);
			state->next_expected_hash = cmd_info->command->prev->digest;
		} else {
			dprintf(9, "no next_expected_hash");
			state->next_expected_hash = NULL;
		}
	} 

	update_state_cmd(state, cmd_info->command);
	return 1;
}

/* compare hash as its logged (ascii) to hash as its stored (binary) */
int cmp_hash(char *asc_hash, unsigned char *bin_hash) 
{
	char asc_hash2[44];

	sprintf(asc_hash2, "%02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x", 
			bin_hash[0], bin_hash[1], bin_hash[2], bin_hash[3],
			bin_hash[4], bin_hash[5], bin_hash[6], bin_hash[7],
			bin_hash[8], bin_hash[9], bin_hash[10], bin_hash[11],
			bin_hash[12], bin_hash[13], bin_hash[14], bin_hash[15],
			bin_hash[16], bin_hash[17], bin_hash[18],
			bin_hash[19]);

	dprintf(9, "hash1 %s hash2 %s", asc_hash, asc_hash2);
	if ( strncmp(asc_hash, asc_hash2, HASH_LENGTH) == 0 ) {
		dprintf(8, "cmp_hash: match found");
		return 1;
	}

	dprintf(8, "cmp_hash: no match found");
	return 0;
}

/* find pointer to cmd_def and to cmd itself for specified hash and hostname:
   1) search config for host with specified hostname
   2) search config->command foreach in  host->allowed_command_set looking for hash
   return current config->command and current command in config->command->command_list
 */
cmd_info_t *find_cmd_info(config_t *config, state_t *state, char *hash)
{
	cmd_info_t *cmd_info;
	host_cfg_t *h;	
	char	   *hostname;

	hostname = state->host;
	
	cmd_info = NULL;
	
	h = get_host(config->host, hostname);
	if ( h == NULL )
		return NULL;
	dprintf(9, "find_cmd_info with hash %s", hash);
	dprintf(9, "calling find_hash %lx, %lx, %lx", h->allowed_command_set, config->command, hash);
	cmd_info = find_hash(h->allowed_command_set, config->command, hash, state);
	if ( cmd_info != NULL ) {
		dprintf(8, "find_cmd_info: cmd_info found");
		return cmd_info;
	}

	dprintf(8, "find_cmd_info: nothing found");
	return NULL;
}

/* search cmd_def foreach in allowed_command_set looking for hash */
cmd_info_t *find_hash(tsw_cmd_t *allowed_cmd_set, cmd_def_t *cmd_def, char *hash, state_t *state)
{
	cmd_info_t *r;
	tsw_cmd_t  *tmp_c; 		/* allowed command */
	cmd_def_t  *tmp_cd;		/* command definition */

	dprintf(8, "find_hash with hash %s", hash);

	r = NULL;

	for (tmp_c = allowed_cmd_set; tmp_c != NULL; tmp_c = tmp_c->next) {
		for ( tmp_cd = cmd_def; tmp_cd != NULL; tmp_cd = tmp_cd->next ) {
			if ( strncmp(tmp_c->command, tmp_cd->name, MAX_LINE_LENGTH) == 0 ) {
				dprintf(9, "searching command definition %s (%s)", tmp_cd->name, tmp_c->command);
				r = find_hash_helper(tmp_cd, state, hash);
				if ( r != NULL )
					return r;
			}
		}
	}
	dprintf(8, "find_hash: nothing found");
	return NULL;
}

/* return > 0 if 'now' is within cmd_def's time specs 
   return 0 if not
 */
int time_ok(tsw_time_t *tsw_time)
{
	time_t 	clock;
	char   	tmp[10]; /* hold strftime strings */
	char	tmp1[2]; 
	int	weekday;
	int	hour;
	int	minute;
	int	second;
	tsw_time_t	*tmp_time;
	se_time_t	*se;

	tzset();
	clock = time(NULL);
	
	/* kind of ugly, but avoids subsequent calls to strftime */
	strftime(tmp, 100, "%w %H %M %S", localtime(&clock));
	tmp1[0] = tmp[0]; tmp1[1] = '\0';
	weekday = atoi(tmp1);
	tmp1[0] = tmp[2]; tmp1[1] = tmp[3]; tmp1[2] = '\0';
	hour = atoi(tmp1);
	tmp1[0] = tmp[5]; tmp1[1] = tmp[6]; tmp1[2] = '\0';
	minute = atoi(tmp1);
	tmp1[0] = tmp[8]; tmp1[1] = tmp[9]; tmp1[2] = '\0';
	second = atoi(tmp1);

	dprintf(7, "time_ok: localtime %s\n", tmp);
	dprintf(7, "time_ok: weekday %d hour %d minute %d second %d\n", weekday, hour, minute, second);
	fflush(stdout);

	for ( tmp_time = tsw_time; tmp_time != NULL; tmp_time = tmp_time->next ) {
		for ( se = tmp_time->se; se != NULL; se = se->next ) {
			if ( se->weekday[weekday] == SET && 		/* weekday */
					( hour > se->start.hour ||	/* start time */
					  ( ( hour == se->start.hour && 
					      minute > se->start.minute ) ||
					    ( hour == se->start.hour && 
					      minute == se->start.minute && 
					      second >= se->start.second ) 
					  )  
					) &&
					( hour < se->end.hour ||	/* end time */
					  ( ( hour == se->end.hour &&
					      minute < se->end.minute ) ||
					    ( hour == se->start.hour &&
					      minute == se->end.minute &&
					      second <= se->end.second )
					  )
					)
			   ) {
				dprintf(8, "time_ok: within %d %d %d - %d %d %d",
						se->start.hour, se->start.minute, se->start.second,
						se->end.hour, se->end.minute, se->end.second);
				return 1;
			}
		}
	}

	return 0;
}

/* add address of command into state's array for found addresses 
   update number of hashes encountered
 */
int update_state_cmd(state_t *state, tsw_cmd_t *command)
{
	state->cmd_addr = add_cmd_addr_front(state->cmd_addr, new_cmd_addr());
	state->cmd_addr->addr = command;
	state->hashes++;

	return 1;
}

cmd_addr_t *new_cmd_addr(void)
{
	cmd_addr_t *newp;

	newp = (cmd_addr_t *) ecalloc(1, sizeof(cmd_addr_t));
	newp->next = NULL;
	return newp;
}

cmd_addr_t *add_cmd_addr_front(cmd_addr_t *listp, cmd_addr_t *newp)
{
	newp->next = listp;
	return newp;
}

se_time_t *add_se_time_front(se_time_t *listp, se_time_t *newp)
{
	newp->next = listp;
	return newp;
}

se_time_t *new_se_time(void)
{
	se_time_t *newp;

	newp = (se_time_t *) ecalloc(1, sizeof(se_time_t));
	newp->next = NULL;
	return newp;
}

/* check whether command is within list of command_addr 
   returns 0 if not found
 */
int check_cmd_addr(cmd_addr_t *cmd_addr, tsw_cmd_t *command)
{
	cmd_addr_t *tmp;
	for ( tmp = cmd_addr; tmp != NULL; tmp = tmp->next ) 
		if ( tmp->addr == command )
			return 1;

	return 0;
}

/* finds cmd to specified hash within specified cmd_def 
   returns it if it's not already listed in the specified state
 */
cmd_info_t *find_hash_helper(cmd_def_t *cmd_def, state_t *state, char *hash)
{
	cmd_info_t *r;
	tsw_cmd_t *tmp_cmd;
	
	r = NULL;

	/* search for a exact match */
	for ( tmp_cmd = cmd_def->command_list; tmp_cmd != NULL; tmp_cmd = tmp_cmd->next ) {
		dprintf(10, "checking command %s ", tmp_cmd->command);
		/* check whether this is an already encountered digest */
		if ( check_cmd_addr(state->cmd_addr, tmp_cmd) ) {
			dprintf(11, "command used before!");
			continue;
		}
		dprintf(11, "calling cmp_hash with hash %s and tmp_cmd->digest at %lx", hash, tmp_cmd->digest);
		if ( cmp_hash(hash, tmp_cmd->digest) ) {
			r = ecalloc(1, sizeof(r));
			r->cmd_def = cmd_def;
			r->command = tmp_cmd;
			dprintf(12, "find_hash_helper: hash found");
			return r;
		}
	}
	/* search for an "ANY" command type match */
	dprintf(8, "find_hash_helper: looking for an any command");
	for ( tmp_cmd = cmd_def->command_list; tmp_cmd != NULL; tmp_cmd = tmp_cmd->next ) {
		if ( check_cmd_addr(state->cmd_addr, tmp_cmd) ) {
			continue;
		}
		dprintf(8, "find_hash_helper: cmd->mode is %d", tmp_cmd->mode);
		if ( tmp_cmd->mode == ANY_CMD ) {
			r = ecalloc(1, sizeof(r));
			r->cmd_def = cmd_def;
			r->command = tmp_cmd;
			dprintf(8, "find_hash_helper: an any command found");
			return r;
		}
	}

	return r;
}

/* 
 * preprocess time specification given in configuration file into
 * a format that is easier to check at runtime
 */
int preprocess_time_spec(config_t *config)
{
	cmd_def_t 	*tmp_c;
	host_cfg_t	*tmp_h;
	tsw_time_t	*tmp_t;
	int		i;

	if ( config == NULL )
		return 0;
	dprintf(8, "preprocess_time_spec: processing time");

	/* command definitions */
	for ( tmp_c = config->command; tmp_c != NULL; tmp_c = tmp_c->next ) {
		dprintf(9, "preprocess_time_spec: processing time for command %s", tmp_c->name);
		for ( tmp_t = tmp_c->time; tmp_t != NULL; tmp_t = tmp_t->next ) {
			for ( i = 0; i < NR_OF_WEEKDAYS; i++ ) {
				if ( tmp_t->type == PERIODIC && tmp_t->weekday[i] == SET )
					preprocess_time(tmp_t, i);
			}
		}
	}

	/* host definitions (automated time) */
	for ( tmp_h = config->host; tmp_h != NULL; tmp_h = tmp_h->next ) {
		dprintf(9, "preprocess_time_spec: processing time for host %s", tmp_h->name);
		for ( tmp_t = tmp_h->automated_time; tmp_t != NULL; tmp_t = tmp_t->next ) {
			for ( i = 0; i < NR_OF_WEEKDAYS; i++ ) {
				if ( tmp_t->type == PERIODIC && tmp_t->weekday[i] == SET )
					preprocess_time(tmp_t, i);
			}
		}
	}

	dprintf(8, "preprocess_time_spec: done with processing time");
	return 1;
}

int preprocess_time(tsw_time_t *time_spec, int day)
{
	se_time_t	*se;
	long		delta;
	long		midnight;
	int		d, h, m;

	dprintf(9, "preprocess_time: processing time for day %d", day);
	
	se = ecalloc(1, sizeof(se_time_t));
	se->start.hour = time_spec->start_hour;
	se->start.minute = time_spec->start_minute;
	se->start.second = 0;
	se->weekday[day] = SET;
	d = day;

	delta = calc_delta(time_spec);
	delta %= WEEK_SECONDS;	/* mod 1 week */

	while ( delta > 0 ) {
		dprintf(10, "preprocess_time: delta %ld", delta);

		midnight = (24 - se->start.hour)*HOUR_SECONDS
			- se->start.minute*MINUTE_SECONDS
			- se->start.second;

		add_start(time_spec, d, se);
		if ( delta > midnight ) {
			add_end(time_spec, d, 23, 59, 59);
			delta -= midnight;
			se->start.hour = 0;
			se->start.minute = 0;
			se->start.second = 0;
			d++;
			d %= 7;
		} else {
			h = floor(delta/(HOUR_SECONDS));
			se->end.hour = se->start.hour + h;
			add_end(time_spec, d, se->end.hour, se->end.minute, se->end.second);
			delta -= h*HOUR_SECONDS;
			
			m = floor(delta/MINUTE_SECONDS);
			se->end.minute = se->start.minute + m;
			delta -= m*MINUTE_SECONDS;

			se->end.second = se->start.second + delta;
			delta = 0;

			/* fixup */
			if ( se->end.second > 59 ) {
				se->end.second -= 60;
				se->end.minute++;
				if( se->end.minute > 59 ) {
					se->end.minute -= 60;
					se->end.hour++;
				}
			}

		}
	}

	free(se);
	return 0;
}

long calc_delta(tsw_time_t *time_spec)
{
	long		delta; /* in seconds */

	delta = 0;
	if ( &time_spec->cont_day != NULL ) 
		delta += time_spec->cont_day*DAY_SECONDS;
	dprintf(13, "calc_delta %ld (%d %d)", delta, time_spec->cont_day, DAY_SECONDS);
	if ( &time_spec->cont_hour != NULL ) 
		delta += time_spec->cont_hour*HOUR_SECONDS;
	dprintf(13, "calc_delta %ld (%d %d)", delta, time_spec->cont_hour, HOUR_SECONDS);
	if ( &time_spec->cont_minute != NULL ) 
		delta += time_spec->cont_minute*MINUTE_SECONDS;
	dprintf(13, "calc_delta %ld", delta);
	if ( &time_spec->cont_second != NULL ) 
		delta += time_spec->cont_second;
	dprintf(13, "calc_delta %ld", delta);
	return delta;
}

int maint_pass(config_t *config)
{
	int 		g_flag = 0;
	int		i;
	host_cfg_t	*tmp_h;
	char		*p = NULL;
	char		*s;
	const char	*pr = "Passphrase for ";
	SHA_CTX 	sha;

	for ( tmp_h = config->host; tmp_h != NULL; tmp_h = tmp_h->next ) {
		if ( tmp_h->admin_mode == INTERACTIVE ) {
			dprintf(7, "maint_pass: interactive password for %s", tmp_h->name);
			s = ecalloc(strlen(tmp_h->name)+strlen(pr)+2, sizeof(char));
			sprintf(s, "%s%s: ", pr, tmp_h->name);
			dprintf(7, "now reading passphrase");
			while ( p == NULL )
				p = read_pass(s);
			dprintf(7, "passphrase is %s", p);
			free(s);
			SHAInit(&sha);
			dprintf(7, "strlen %s is %d", p, (int *) strlen(p));
			SHAUpdate(&sha, p, strlen(p));
			SHAFinal(tmp_h->admin_digest, &sha);
			/* clear the passphrase */
			for ( i=0; i < strlen(p); i++ )
				p[0] = '\0';
			free(p);
		} else if ( tmp_h->admin_mode == GLOBAL ) {
			dprintf(7, "maint_pass: global password for %s", tmp_h->name);
			g_flag++;
		}
	}

	if ( g_flag ) {
		while ( p == NULL )
			p = read_pass("Global passphrase: ");
		SHAInit(&sha);
		dprintf(7, "strlen %s is %d", p, (int *) strlen(p));
		SHAUpdate(&sha, p, strlen(p));
		SHAFinal(config->admin_digest, &sha);
		/* clear the passphrase */
		for ( i=0; i < strlen(p); i++ )
			p[0] = '\0';
		free(p);
	}
	
	return 0;
}

char *read_pass(char *s)
{
	char *passphrase,*p;
	p = getpass(s);
	passphrase = (char *) calloc(strlen(p)+1, sizeof(char));
	strncpy(passphrase, p, strlen(p)+1);
	dprintf(7, "passphrase is %s", passphrase);
	dprintf(8, "passphrase length is %d", strlen(passphrase));
	if(strlen(passphrase) > 0)
		return passphrase;
	return NULL;
}

int automated_action(state_t *state, host_cfg_t *host_cfg)
{
	state->automated_action++;
	dprintf(2, "automated_action: for host %s", host_cfg->name);

	/* are we within an automated action for this state? */
	/* (no locking ...) */
	if ( state->automated_action > 1 )
		return 1;

	if ( time_ok(host_cfg->automated_time) ) {
		dprintf(3, "automated_action: within automated time");
		/* let's rock */
		switch(fork()) {
			case -1:
				/* fork failed: TODO do something! */
				return (-1);
			case 0:
				/* child */
				dprintf(4, "automated_action: executing command '%s'", host_cfg->automated_action);
				system(host_cfg->automated_action);
				log_automated_action(host_cfg);
				exit(0);
			default:
				break;


		}
		return 2;
	} else 
		dprintf(3, "automated_action: not within automated time");

	return 0;
}

int log_alert(host_cfg_t *host_cfg)
{
	dprintf(2, "log_alert: for host %s", host_cfg->name);
	openlog(host_cfg->syslog_tag, LOG_CONS | LOG_NDELAY, host_cfg->facility);
	syslog(host_cfg->priority, host_cfg->syslog_message);
	closelog();
	return 0;
}

int log_automated_action(host_cfg_t *host_cfg)
{
	dprintf(2, "log_alert: for host %s", host_cfg->name);
	openlog(host_cfg->syslog_tag, LOG_CONS | LOG_NDELAY, host_cfg->facility);
	syslog(host_cfg->priority, "automated action executed");
	closelog();
	return 0;
}

int log_ok(host_cfg_t *host_cfg)
{
	dprintf(2, "log_ok: for host %s", host_cfg->name);
	openlog(host_cfg->syslog_tag, LOG_CONS | LOG_NDELAY, host_cfg->facility);
	syslog(LOG_DEBUG, "expected command (ok).");
	closelog();
	return 0;
}

int add_start(tsw_time_t *time_spec, int day, se_time_t *se)
{
	dprintf(11, "add_start: day %d %d:%d:%d", day, se->start.hour, se->start.minute, se->start.second);
	time_spec->se = add_se_time_front(time_spec->se, new_se_time());
	time_spec->se->start.hour = se->start.hour;
	time_spec->se->start.minute = se->start.minute;
	time_spec->se->start.second = se->start.second;
	time_spec->se->weekday[day] = SET;
	dprintf(12, "add_start: time_spec->se %lx day %d %d:%d:%d weekday %d", time_spec->se, day, time_spec->se->start.hour, time_spec->se->start.minute, time_spec->se->start.second, time_spec->se->weekday[day]);
	return 0;
}

int add_end(tsw_time_t *time_spec, int day, int hour, int minute, int second)
{
	dprintf(11, "add_end: day %d %d:%d:%d", day, hour, minute, second);
	assert( time_spec->se->weekday[day] == SET );
	time_spec->se->end.hour = hour;
	time_spec->se->end.minute = minute;
	time_spec->se->end.second = second;
	return 0;
}

int preprocess_syslog_spec(config_t *config)
{
	host_cfg_t	*tmp_h;

	for ( tmp_h = config->host; tmp_h != NULL; tmp_h = tmp_h->next ) {

		dprintf(1, "preprocess_syslog_spec: %s", tmp_h->name);
		/* facility munging */
		if ( strncasecmp(tmp_h->syslog_facility, "user", MAX_LINE_LENGTH) == 0 ) {
			tmp_h->facility = LOG_USER;
		} else if ( strncasecmp(tmp_h->syslog_facility, "local", 5) == 0 ) {
			switch(tmp_h->syslog_facility[5]) {
				case '0':
					tmp_h->facility = LOG_LOCAL0;
				case '1':
					tmp_h->facility = LOG_LOCAL1;
				case '2':
					tmp_h->facility = LOG_LOCAL2;
				case '3':
					tmp_h->facility = LOG_LOCAL3;
				case '4':
					tmp_h->facility = LOG_LOCAL4;
				case '5':
					tmp_h->facility = LOG_LOCAL5;
				case '6':
					tmp_h->facility = LOG_LOCAL6;
				case '7':
					tmp_h->facility = LOG_LOCAL7;
				default:
					tmp_h->facility = LOG_DAEMON;
			}

		} else {
			tmp_h->facility = LOG_DAEMON;
		}

		/* priority munging */
		dprintf(1, "preprocess_syslog_spec: priority munging");
		if (strncasecmp(tmp_h->syslog_priority, "debug", MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: debug %d", LOG_DEBUG);
			tmp_h->priority = LOG_DEBUG;
		} else if (strncasecmp(tmp_h->syslog_priority, "notice", MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: error %d", LOG_NOTICE);
			tmp_h->priority = LOG_NOTICE;
		} else if (strncasecmp(tmp_h->syslog_priority, "warning", MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: error %d", LOG_WARNING);
			tmp_h->priority = LOG_WARNING;
		} else if (strncasecmp(tmp_h->syslog_priority, "error", MAX_LINE_LENGTH) == 0) {
			tmp_h->priority = LOG_ERR;
			dprintf(1, "priority: error %d", LOG_ERR);
		} else if (strncasecmp(tmp_h->syslog_priority, "critical", MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: error %d", LOG_CRIT);
			tmp_h->priority = LOG_CRIT;
		} else if (strncasecmp(tmp_h->syslog_priority, "alert", MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: error %d", LOG_ALERT);
			tmp_h->priority = LOG_ALERT;
		} else if (strncasecmp(tmp_h->syslog_priority, "emergency", MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: emergency %d", LOG_EMERG);
			tmp_h->priority = LOG_EMERG;
		} else {
			dprintf(1, "priority: crit %d", LOG_CRIT);
			tmp_h->priority = LOG_CRIT;
		}

	}

	return 0;
}

struct cmd_line_t *cmd_parse(int argc, char **argv)
{
	int c;
	extern char *optarg;
	extern int optind;
	struct cmd_line_t *cmd_line_opts;

	dprintf(2, "cmd_parse: parsing command line arguments");
	cmd_line_opts = (struct cmd_line_t *) calloc(1, sizeof(struct cmd_line_t));

	while ((c = getopt(argc, argv, "c:d:")) != -1 ) {
		switch (c) {
			case 'c':
				cmd_line_opts->file = estrdup(optarg);
				dprintf(3, "cmd_parse: file is %s", cmd_line_opts->file);
				break;
			case 'd':
				cmd_line_opts->debug = atoi(optarg);
				dprintf(3, "cmd_parse: debug is %d", cmd_line_opts->debug);
				break;
		}
	}
	for (; optind < argc; optind++ ) {
		printf("extra argument: %s\n", argv[optind]);
	}

	return cmd_line_opts;
}

void usage(void)
{
	printf("%s\n", progname());
	printf("usage: %s -c file [ -d int ]\n", progname());
	printf(" file\t\tconfiguration file\n");
	printf(" -d int\t\tprint debug information (1 < int < 16)\n");
}
